#!/usr/bin/perl 
# repeater.perl
#
# Analysis of repeats in a database using a negative and a positive 
# profile. Allows two modes. In mode calibrate it tries the profile
# against a database in order to find the EVD distribution.
# In mode search, it extracts repeats applying the EVD parameters.
#
# Use as: repeater.perl -neg negative profile -pos positive profile
#         -d fasta database -mode [calibrate|search] [-thr threshold]
#         [-for EVD parameters file] 
#
# Example: repeater.perl -neg heat_neg.prf -pos heat_pos.prf 
#          -d rswiss_5 -mode calibrate
#
# Options: -neg negative profile
#          -pos positive profile
#          -d database in fasta format
#          -mode indicates wether a calibration is performed or a
#           search on a database.
#          -for (only in search mode) EVD fitted parameters as
#           generated by this program in calibrate mode.
#          -thr (only in search mode) Threshold in E-value 
#           for repeat annotation.
#
# Input: - profiles (as generated by pairwise)
#        - database (in fasta format)
#        - EVD fitted parameters (only in search mode)
#
# Output: - EVD fitting parameters (only in calibrate mode)
#         - parameters about fragment hits (only in search mode)
#
# Needs:  - swise
#
# Version 1.1 : There is a bug in searchwise. All hits are reported
#               in the *.out file but the las alignment is not reported
#               in the *_aln.out file. 
#               Till it gets fixed (Julie Thompson) I am adding a dummy 
#               sequence to the end of any fasta file to be used.
#
# Version 1.2 : Searchwise requires no space before identifier in a fasta
#               file. Names get messed up if you do it.
#
# Version 1.3 : But names get messed anyway sometimes (e.g. with nrdb)
#               They go to the third line of the output. Detect that and correct.
#
# Version 1.4 : Searchwise cannot deal either with dots in the identifiers
#               (e.g. in wormpep). Substitute them by underscores.
#
# Version 2.0 : Include column of protein length, so that this property can
#               be optionally used later for selection. This change affects
#               select.perl (and any other programs reading *.par files)
#
# Version 2.1 : Include option to keep the hit files [-keep] for
# (Jun 99)      statistical purposes. Only valid in calibrate.
#
# Version 3.0 : Adapted for server. Perl header. 
# (June 99)                         Output in screen in search mode.
#                                   Add booby/warbler switch for swise source
# Version 3.1 : Repeater uses only swise and one matrix. It does not need
#               any source'ing or com files to run. Now swise (1.7) is directly
#               executed (location of the source is therefore needed somewhere)
#               as well as the location and name of the matrix to use (fixed).
#
#               In the search mode, the
#               names of temporal files now deppend on the name of the fasta
#               file in order to avoid clashes of multiple runs of the script.
#
# Version 3.2 : Error readind profile name in FRG_ALN. Name can be in a 
#               directory. Directories names are thrown away.
# Version 3.3 : swise bug. -q output was not working unless -o option is
#               used.
# Version 3.4 : But in calibrate the output was going also on screen!
#
# Version 3.5 : Added a more restrictive termination condition for search mode
# 7 Dec 2000    in order to fix never ending swise bug
#
# Miguel A. Andrade - Sept 98
# Repeats project
#########################################################################



$PROG_ID = 'repeater.perl 3.4';

# $system = 'tau';
# $system = 'warbler';
# $system = 'booby';
$system = 'mohamed';
# $system = 'miguel_mainz';

if ($system eq 'booby'){
  # for the server
  $swise_source = '/booby1/home/andrade/public_html/rep/swise';
  $matrix = '/booby1/home/andrade/public_html/rep/gon250.bla';

} elsif ($system eq 'tau') {
  # $swise_source = '/home/seqanal/packages/prepare/wisetools.com';
  $swise_source = '/home/seqanal/packages/OSF1/swise';
  # $matrix = '/usr/local/lib/wisetools.1.6.1/data/gon250.bla';
  $matrix = '/home/seqanal/packages/wisetools/OSF1setup/gon250.bla';

} elsif ($system eq 'warbler') {
  # $swise_source = '/home/seqanal/packages/prepare/wisetools.com';
  $swise_source = '/home/seqanal/packages/IRIX/swise';
  # $matrix = '/usr/local/lib/wisetools.1.6.1/data/gon250.bla';
  $swise_source = '/home/seqanal/packages/wisetools/IRIXsetup/gon250.bla';
}
 elsif ($system eq 'mohamed') {
  $swise_source = '/Users/mohamed/Sites/rep/swise';
  $matrix = '/Users/mohamed/Sites/rep/gon250.bla';
}
 elsif ($system eq 'miguel_mainz') {
  $swise_source = '/home/munoz/public_html/cgi-bin/rep/swise';
  $matrix = '/home/munoz/public_html/cgi-bin/rep/gon250.bla';
}



# get the working directory
$working_dir = `pwd`;
chop($working_dir);
$working_dir .= '/';


# read command line

if ($#ARGV < 5){
  print "Use as: repeater.perl -neg negative profile -pos positive profile\n";
  print "-d fasta database -mode [calibrate|search] [-for EVD parameters file]\n";
  die "[-thr e-value threshold] [-keep]\n";
}

$THRESHOLD = '';
$NEGATIVE = '';
$POSITIVE = '';
$ORIGINAL_DB = '';
$mode = '';
$EVD_IN = '';
$KEEP = 0;

for ($i = 0; $i <= $#ARGV; $i++){
  if ($ARGV[$i] eq '-neg') {
    $NEGATIVE = $ARGV[$i+1];
    $i++;
    next;
  }
  if ($ARGV[$i] eq '-pos') {
    $POSITIVE = $ARGV[$i+1];
    $i++;
    next;
  }
  if ($ARGV[$i] eq '-d') {
    $ORIGINAL_DB = $ARGV[$i+1];
    $i++;
    next;
  }
  if ($ARGV[$i] eq '-mode') {
    $mode = $ARGV[$i+1];
    $i++;
    next;
  }
  if ($ARGV[$i] eq '-for') {
    $EVD_IN = $ARGV[$i+1];
    $i++;
    next;
  }
  if ($ARGV[$i] eq '-thr') {
    $THRESHOLD = $ARGV[$i+1];
    $i++;
    next;
  }
  if ($ARGV[$i] eq '-keep') {
    $KEEP = 1;
    next;
  }
}


if ($NEGATIVE eq '') {
  die "Exit: Must provide a negative profile\n";
}
if ($POSITIVE eq '') {
  die "Exit: Must provide a positive profile\n";
}
if ($ORIGINAL_DB eq '') {
  die "Exit: Must provide a database file\n";
}
if (($mode ne 'calibrate') && ($mode ne 'search')) {
  die "Exit: Must provide a mode, either calibrate or search\n";
}
if (($mode eq 'search') && ($EVD_IN eq '')) {
  die "Exit: In search mode you must provide fitted EVD parameters\n";
}
if (($mode eq 'search') && ($KEEP == 1)) {
  die "Exit: keep option not possible in search mode\n";
}


# define constants

# swise parameters
$max_alignments = 20000;

# maximum fraction of masking in a fragment in order to 
# consider it
$max_mask = 0.66;


# mode-dependent parameters

if ($mode eq 'calibrate') {
  # minimum number of sequences left for continuing cycles
  $min_sequences = 500;

  # swise parameters
  $max_e_value = 10000;

} else {
  # e-value threshold
  if ($THRESHOLD eq '') {
    # use default
    $e_threshold = 1e-4;
  } else {
    $e_threshold = $THRESHOLD;
  }

  # minimum number of sequences left for continuing cycles
  $min_sequences = 1;

  # swise parameters
  $max_e_value = 10000;

}

# searchwise bug associated:
# dummy sequence is added to any fasta file to be analyzed
$add = ">dummy\n";
$add .= "AAAAA\n";




# analyse negative profile

$gap_opening_neg = '';
$gap_extension_neg = '';
$length_neg = '';
$sign_neg = '';
$sequences_neg = '';
$matrix_neg = '';

open(NEGATIVE, $NEGATIVE) or die "negative profile file $NEGATIVE not found\n";
while(<NEGATIVE>) {
  if (/Comparison\s+Matrix/) {
    @words = split(/\s+/);
    $sign_neg = $words[3];
    $matrix_neg = $words[2];
  }
  if (/Length\:\s+(\S+)$/) {
    $length_neg = $1;
  }
  if (/^Sequences\:\s+(\S+)\s+/) {
    $sequences_neg = $1;
  }
  if (/Comparison\s+Matrix/) {
    @words = split(/\s+/);
    $sign = $words[3];
  }
  if (/.*WISE.*GO\:\s+(\S+)\s+/) {
    $gap_opening_neg = $1;
  }
  if (/.*WISE.*GE\:\s+(\S+)\s+/) {
    $gap_extension_neg = $1;
  }
}
close(NEGATIVE);

# check profile sign
if ($sign ne 'Negative') {
  die "Exit: profile $NEGATIVE is not negative but $sign_neg\n";
}

# analyse positive profile

$gap_opening_pos = '';
$gap_extension_pos = '';
$length_pos = '';
$sign_pos = '';
$sequences_pos = '';
$matrix_pos = '';

open(POSITIVE, $POSITIVE) or die "negative profile file $POSITIVE not found\n";
while(<POSITIVE>) {
  if (/Comparison\s+Matrix/) {
    @words = split(/\s+/);
    $sign_pos = $words[3];
    $matrix_pos = $words[2];
  }
  if (/Length\:\s+(\S+)$/) {
    $length_pos = $1;
  }
  if (/^Sequences\:\s+(\S+)\s+/) {
    $sequences_pos = $1;
  }
  if (/Comparison\s+Matrix/) {
    @words = split(/\s+/);
    $sign_pos = $words[3];
  }
  if (/.*WISE.*GO\:\s+(\S+)\s+/) {
    $gap_opening_pos = $1;
  }
  if (/.*WISE.*GE\:\s+(\S+)\s+/) {
    $gap_extension_pos = $1;
  }
}
close(POSITIVE);

# check profile sign
if ($sign_pos ne 'Positive') {
  die "Exit: profile $POSITIVE is not positive but $sign_pos\n";
}



# check that the positive and negative profiles are compatible

if ($length_pos != $length_neg) {
  die "Exit: provided profiles differ on length\n";
}
if ($sequences_pos != $sequences_neg) {
  die "Exit: provided profiles differ on number of sequences\n";
}
if ($matrix_pos != $matrix_neg) {
  die "Exit: provided profiles differ on matrix used\n";
}





if ($mode eq 'calibrate') {

  # extract basic name for calibrate from the negative profile name
  ($basic_name, $extension) = split(/\./, $NEGATIVE);

  # use the name as of previous versions
  $working_name = $working_dir.'repeater';

} else {

  # extract basic name for the search mode from the database file
  # Version 3.1
  ($fasta_name, $extension) = split(/\./, $ORIGINAL_DB);

  $working_name = $working_dir.$fasta_name;
}


# build output and temporary file names
$MASK = $working_name.'_mask.fasta';
$FRG = $working_name.'_frg.fasta';
$NEG_IN = $working_name.'_neg.in';
$POS_IN = $working_name.'_pos.in';
$ALN = $working_name.'_aln.out';
$HIT = $working_name.'_hit.out';
$FRG_ALN = $working_name.'_frg_aln.out';
$DB = $working_name.'_db.fasta';

# searchwise bug associated names
$ADD = $working_name.'_add.in';
$DB_ADD = $working_name.'_db_add.fasta';
$FRG_ADD = $working_name.'_frg_add.fasta';

# extra temporal file
# Version 3.3
$TMP = $working_name.'_tmp.out';



# searchwise bug associated:
# prepare a file with the dummy sequence

open(ADD, ">".$ADD);
print ADD $add;
close(ADD);





if ($mode eq 'calibrate') {


  # calibrate: prepare a file for output on formula E=f(score)
  $EVD_OUT = $basic_name.'.for';
  print "Will write EVD fitted parameters to: $EVD_OUT\n";
  open(EVD_OUT, '>'.$EVD_OUT);
  print EVD_OUT "#rdb generated by $PROG_ID\n";
  print EVD_OUT "cycle\tN\ta\tb\n";
  print EVD_OUT "3S\t3S\t10S\t10S\n";

} else {
  
  # search: prepare a file for fragment parameters (rdb)
  # $REP_PAR = $basic_name.'.par';
  # print "Will write aligned fragment parameters to: $REP_PAR\n";
  # open(REP_PAR, '>'.$REP_PAR);
  print "# rdb generated by $PROG_ID\n";
  print "id\tprot_length\torder\tevalue\tenext\t";
  print "score\tlength\tp_start\tp_end\t";
  print "seq_start\tseq_end\tprofile\tfragment\n";
  print "60S\t5N\t3N\t10N\t10N\t";
  print "3N\t3N\t3N\t3N\t";
  print "3N\t3N\t100S\t100S\n";


  # read the formula table (cycle, N, a, b) for converting 
  # scores in E-values
  # (rdb header)

  open(EVD_IN, $EVD_IN) or die "$EVD_IN file not found\n";
  $skip_header = 2;
  while (<EVD_IN>) {
    if (/^\#/) {
      next;
    }
    if ($skip_header > 0) {
      $skip_header--;
      next;
    }
    chop;
    @words = split(/\t/);
    $order = $words[0];
    $for_N{$order} = $words[1];
    $for_a{$order} = $words[2];
    $for_b{$order} = $words[3];
  }
  close(EVD_IN);
  
}




###################################################
# Start cycles of profile scan and db masking.
# Stop when there are not enough sequences left.
###################################################

$go_on = 1;
$cycle = 1;
system(`cp $ORIGINAL_DB $DB`);

while ($go_on == 1) {

  if ($cycle > 1) {
    # Copy the masked database as the input one
    system(`cp $MASK $DB`);
  } 

  # searchwise bug associated:
  # add a dummy sequence at the end
  system(`cat $DB $ADD > $DB_ADD`);


  # set the number of alignments to the size of the database

  $number_of_alignments = `grep -c '^>' $DB_ADD`;
  chop($number_of_alignments);
  
  # version 3.4: offer some info for calibrate mode
  if ($mode eq 'calibrate') {
    print "Running cycle $cycle to find $number_of_alignments hits\n";
  }

  # or if more than swise limit, then use swise limit
  if ($number_of_alignments > $max_alignments) {
    $number_of_alignments = $max_alignments;
  }



  ############################################################
  # generate a swise in file for the negative profile search
  ############################################################

  open(NEG_IN, '>'.$NEG_IN);
  print NEG_IN "! Input file from a fasta database\n";
  print NEG_IN "! Generated by repeater.perl\n";
  print NEG_IN "$DB_ADD fasta protein\n";
  close(NEG_IN);


  #################################################################
  # generate the swise com file for negative profile search
  #################################################################
  # Version 3.1 Not anymore!
  # 
  # $NEG_COM = $working_dir.'repeater_neg.com';
  # open(NEG_COM, '>'.$NEG_COM);
  # print NEG_COM "#!/usr/bin/csh\n";
  # print NEG_COM "source $swise_source\n";
  # print NEG_COM "swise -p$NEGATIVE -i$NEG_IN -y$max_e_value \\\n";
  # print NEG_COM "-m$matrix -g$gap_opening_neg \\\n";
  # print NEG_COM "-e$gap_extension_neg -s$number_of_alignments -q$ALN \\\n";
  # print NEG_COM "-r$number_of_alignments -b2000 -a -o$HIT\n";
  # close(NEG_COM);
  # system(`chmod a+x $NEG_COM`);

  # generate command line for calling directly swise
  # Version 3.1
  $call_neg = "$swise_source -p$NEGATIVE -i$NEG_IN -y$max_e_value -m$matrix ";
  $call_neg .= "-g$gap_opening_neg -e$gap_extension_neg -s$number_of_alignments -q$ALN ";
  $call_neg .= "-r$number_of_alignments -b2000 -a -o$HIT";

  # print $call_neg;


  #######################################
  # run swise using the negative profile
  #######################################
  #system(`echo $call_neg > tmp/t.txt`);
  system(`$call_neg`);
  
  
  
  # Version 2.1 keeping hit files in calibrate mode
  # save the hit file in a new file

  if ($KEEP) {
    $SAVE_HIT = $working_dir.'rep_graph_hit.'.$cycle;
    system(`cp $HIT $SAVE_HIT`);
  }





  if ($mode eq 'calibrate') {
    
    ##################################################################
    # get the e-value function of score parameters 
    # from the top score list
    ##################################################################

    open(HIT, $HIT);
    $search_start = 0;

    while(<HIT>) {
      if (/^E\-Values\:/) {
	$search_start = 1;
	next;
      }
      if ($search_start == 1) {
	if (/\=\s(\S+)\s\*\s\(1\s\-\sexp\((\S+)\s\*\sexp\((\S+)/) {
	  print EVD_OUT "$cycle\t$1\t$2\t$3\n";
	}
      }
    }
    close(HIT);
  }




  ########################################
  # get the hit table from swise output
  ########################################

  open(ALN, $ALN);
  undef %s0;
  undef %s1;
  $counter_sub = 0;
  $counter = 0;
  $numeral = 0;
  $found = 0;
  $search_id = 0;

  while(<ALN>){

    if (/^\_(\d+)\_*$/) {
      # found hit numeral
      
      $found = 0;
      $numeral = $1;
      next;
    }

    if (/^Sequence/) {
      
      # get identifier
      chop;
      @words = split(/\W+/);
      $tmp = $words[1];
      $identifier = $tmp;
      $search_id = 2;
      next;
    }

    if ($search_id == 2) {
      $search_id = 1;
      next;
    }

    if ($search_id == 1) {
      # check wether the 2nd line after Sequence is empty or not
      if (/\S/) {
	# the identifier is here (swise bugS in dealing with nrdb ids)

	# cut identifiers with pipes
	chop;
	@words = split(/\s+/);
        $tmp = $words[0];
	if ($tmp =~ /\|/) {
	  @words2 = split(/\|/, $tmp);
	  $identifier = pop(@words2);
	} else {
	  $identifier = $tmp;
	}
      }
      $search_id = 0;

      # in any case substitute dots by doble underscore in the identifier
      $identifier =~ s/\./\_\_/g;


      next;
      
    }

    if (/^Score\s(\d+)\s+E\-Value\s(.*)$/) {
      # get score and E-value
      $score{$identifier} = $1;
      $found++;
      next;
    }
      
    if (/^In original  coordinates\:\s+(\d+)\-\s*(\d+)\s\(profile\)\s+(\d+)\-\s*(\d+)/) {
      # get identity and length of alignment
      $p0{$identifier} = $1;
      $p1{$identifier} = $2;
      $s0{$identifier} = $3;
      $s1{$identifier} = $4;
      $found++;
    }

    if ($found == 2) {
      # successfully found all information required
      $found = 0;
      
      if ($mode eq 'search') {

	if ($for_N{$cycle} ne '') {
          # use loaded repeat dependent parameters
          # for converting score in E-value
	  $e_value{$identifier} =
	    1 - exp($for_a{$cycle} * (exp($for_b{$cycle}*$score{$identifier})));
	  $last_cycle = $cycle;

          # if possible, compute E-value for n+1 diystribution
          # if not, then put the E-value for the E(n+1)
	  $next_cycle = $cycle + 1;
	  if ($for_N{$next_cycle} ne '') {
	    $enext_value{$identifier} =  
	      1 - exp($for_a{$next_cycle} * (exp($for_b{$next_cycle}*$score{$identifier})));
	  } else {
	    $enext_value{$identifier} = $e_value{$identifier};
	  }

	} else {
          # parameters exhausted.
          # Use instead last parameters loaded	
	  $e_value{$identifier} =  
	    1 - exp($for_a{$last_cycle} * (exp($for_b{$last_cycle}*$score{$identifier})));
	      
	  $enext_value{$identifier} = $e_value{$identifier};
	}

	if (($e_value{$identifier} > $e_threshold) && ($enext_value{$identifier} > $e_threshold)) {
          # eliminate the hit
	  $s0{$identifier} = '';
	} else {
	  # the hit is registered
	  $counter++;
	  $counter_sub++;    
	}
      }
	  
      if ($mode eq 'calibrate') {
	$counter++;
	$counter_sub++;

	if ($counter_sub == 500) {
	  $counter_sub = 0;
	  print "$counter of $number_of_alignments possible alignments extracted\n";
	}

      }

    }

  }
  close(ALN);
  $hits = $counter;

  # termination condition
  if ($hits < $min_sequences) {
    $go_on = 0;
    next;
  }

  


  ###############################################################
  # extract the sequences and fragments for which a hit was found
  ###############################################################

  $get_sequence = 0;
  $sequence = '';

  # version 3.5
  # Check that something is put into the fragment file
  $frg_ok = 0;

  open(FRG, '>'.$FRG);
  open(DB, $DB);
  while(<DB>) {

    if (/^\>\s*(\S+)/) {
      
      if ($get_sequence == 1) {
	&extend_fragment;
	$sequence = '';
      }

      # clean nrdb identifiers
      $tmp = $1;
      if ($tmp =~ /\|/) {
	@words2 = split(/\|/, $tmp);
	$identifier = pop(@words2);
      } else {
	$identifier = $tmp;
      }

      # substitute dots by doble underscore in the identifier
      $identifier =~ s/\./\_\_/g;

      # special modifications can be accomplished on the database
      # names by swise. This is a problem. One has to catch them.
      # $identifier =~ s/^swiss\|//;

      if ($s0{$identifier} ne '') {
	$get_sequence = 1;
      } else {
	$get_sequence = 0;
      }
      
    } else {
      if ($get_sequence == 1) {
	chop;
	$sequence .= $_;
      }
    }


  }
  if ($sequence ne '') {
    &extend_fragment;
  }
  close(DB);
  close(FRG);

  # version 3.5
  # More restrictive termination condition for the search mode
  # Terminate if fragment file is empty
  if ($mode eq 'search'){
    if ($frg_ok == 0){
      $go_on = 0;
      next;
    }
  }
      
    
  # searchwise bug associated:
  # add a dummy sequence at the end
  system(`cat $FRG $ADD > $FRG_ADD`);

  
  ############################################################
  # generate a swise in file for the positive profile search
  # on the fragment database
  ############################################################

  open(POS_IN, '>'.$POS_IN);
  print POS_IN "! Input file from a fasta database\n";
  print POS_IN "! Generated by repeater.perl\n";
  print POS_IN "$FRG_ADD fasta protein\n";
  close(POS_IN);


  #################################################################
  # generate the swise com file for positive profile search
  #################################################################
  # Not in version 3.1!

  # $POS_COM = $working_dir.'repeater_pos.com';
  # open(POS_COM, '>'.$POS_COM);
  # print POS_COM "#!/usr/bin/csh\n";
  # print POS_COM "source $swise_source\n";
  # print POS_COM "swise -p$POSITIVE -i$POS_IN -y$max_e_value \\\n";
  # print POS_COM "-m$matrix -g$gap_opening_pos \\\n";
  # print POS_COM "-e$gap_extension_pos -s$number_of_alignments -q$FRG_ALN \\\n";
  # print POS_COM "-r$number_of_alignments -b2000 -a\n";
  # close(POS_COM);
  # system(`chmod a+x $POS_COM`);

  # generate a direct call to swise
  $call_pos = "$swise_source -p$POSITIVE -i$POS_IN -y$max_e_value ";
  $call_pos .= "-m$matrix -g$gap_opening_pos -e$gap_extension_pos ";
  # Version 3.3 added option -o, needed for -q to work in the server
  # Strange!
  $call_pos .= "-s$number_of_alignments -q$FRG_ALN -o$TMP ";
  $call_pos .= "-r$number_of_alignments -b2000 -a";

  # print $call_pos;


  #################################################################
  # run swise using the positive profile on the fragment database
  #################################################################
  `$call_pos`;


  ################################################################
  # get the hit table from swise output of fragment analysis
  # with positive profile
  ################################################################

  open(FRG_ALN, $FRG_ALN);
  $s_name = 0;
  $s_rest = 0;
  $counter_sub = 0;
  $counter = 0;
  $numeral = 0;
  $get_next_sequence = 0;
  $found = 0;

  while(<FRG_ALN>){
    
    if ($profile_name eq '') {
      if (/^Using profile in file (\S+)$/) {
	$profile_name = $1;

	# Version 3.2
	# Eliminate possible directory path
	# (This may be necessary deppending on the swise version used)
	if ($profile_name =~ /\//) {
	  @words = split(/\//, $profile_name);
	  $profile_name = pop(@words);
	}

        # correct for length limitation
	if (length($profile_name) > 10) {
	  $tmp = $profile_name;
	  $profile_name = substr($tmp, 0, 10);
	}
      }
      next;
    }

    if (/^$profile_name/) {
      @words = split(/\W*\s+\W*/);
      $profile .= pop(@words);
      $get_next_sequence = 1;
      next;
    }

    if ($get_next_sequence == 1) {
      @words = split(/\W*\s+\W*/);
      $hit .= pop(@words);
      $get_next_sequence = 0;
      next;
    }

    if (/^\_(\d+)\_*$/) {
      # found hit numeral

      $numeral = $1;
      $s_name = 3;
      $profile{$identifier} = $profile;
      $hit{$identifier} = $hit;

      $hit = '';
      $profile = '';
      next;
    }

    if (/^Sequence/) {

      # get identifier
      chop;
      @words = split(/\W+/);
      $identifier = $words[1];
      next;
    }

    if (/^In original\s+coordinates\:\s+(\d+)\-\s*(\d+)\s\(profile\)\s+(\d+)\-\s*(\d+)/) {
      # get identity and length of alignment
      $p0_frg{$identifier} = $1;
      $p1_frg{$identifier} = $2;
      $s0_frg{$identifier} = $3;
      $s1_frg{$identifier} = $4;

      $found = 1;
	
      if ($found == 1) {
	# successfully found all information required
	$counter++;
	$counter_sub++;
	if ($counter_sub == 500) {
	  $counter_sub = 0;
	  print "$counter of $hits fragment alignments extracted\n";
	}
	
       
      }
    }

  }

  # annotate the last one
  $profile{$identifier} = $profile;
  $hit{$identifier} = $hit;

  close(FRG_ALN);
  $frg_hits = $counter;


  ###############################################################
  # extract the sequences for which a hit was found mask,
  # annotate parameters and write alignment
  ###############################################################

  $get_sequence = 0;
  $sequence = '';

  open(MASK, '>'.$MASK);
  open(DB, $DB);
  while(<DB>) {

    if (/^\>\s*(\S+)/) {
      if ($get_sequence == 1) {
	&mask_sequence;
	$sequence = '';
      }

      # clean nrdb identifiers
      $tmp = $1;
      if ($tmp =~ /\|/) {
	@words2 = split(/\|/, $tmp);
	$identifier = pop(@words2);
      } else {
	$identifier = $tmp;
      }

      # substitute dots by doble underscore in the identifier
      $identifier =~ s/\./\_\_/g;


      # special modifications can be accomplished on the database
      # names by swise. This is a problem. One has to catch them.
      # $identifier =~ s/^swiss\|//;

      if ($s0{$identifier} ne '') {
	$get_sequence = 1;
      } else {
	$get_sequence = 0;
      }
      
    } else {
      if ($get_sequence == 1) {
	chop;
	$sequence .= $_;
      }
    }


  }
  if ($sequence ne '') {
    &mask_sequence;
  }
  
  close(DB);
  close(MASK);


  # finished one cycle
  # version 3.4: offer some info in mode calibrate
  if ($mode eq 'calibrate') {
    print "cycle $cycle finished. Progressing to next.\n";
  }

  $cycle++;

} # end of one masking cycle



# clean directory

close(EVD_OUT);







########################
# end
########################






# extends a hit according to position of the hit and profile lenght

sub extend_fragment {
  
  $sequence_length = length($sequence);

  # if this is the first cycle, annotate length of the sequence
  # Version 2.0
  if ($cycle == 1) {
    $prot_length{$identifier} = $sequence_length;
  }

  $start = $s0{$identifier};
  $stop = $s1{$identifier};
  
  # extend the fragment around the hit
  # by length of the profile minus length of the hit
  $extend_down = int(($p0{$identifier} - 1) * 1.5);
  $extend_up = int(($length_neg - $p1{$identifier}) * 1.5);

  $start -= $extend_down;
  $stop += $extend_up;

  if ($stop > $sequence_length) {
    $stop = $sequence_length;
  }
  if ($start < 1) {
    $start = 1;
  }
  $fragment_length = $stop - $start;

  # convert notation of sequences to string notation
  $start--;
  $stop--;
  $fragment_length++;

  $fragment = substr($sequence, $start, $fragment_length);

  # check that the constructed fragment is not already
  # heavily masked

  $string = $fragment;
  $count = ($string =~ tr/X//);
  $masked_fraction = $count / $fragment_length;

  if ($masked_fraction < $max_mask) {
    # annotate offset for later report of the position of the hit
    $offset{$identifier} = $start;
      
    print FRG ">$identifier\n";
    print FRG "$fragment\n";

    # version 3.5
    # Something was put into the fragment file
    $frg_ok = 1;

  } else {
    
    # forget about this sequence
    $s0{$identifier} = '';
    $s1{$identifier} = '';
  }


}


# masks a sequence according to a hit to it

sub mask_sequence{


  # $start = $s0_frg{$identifier} - 1;
  # $stop = $s1_frg{$identifier} - 1;
  $start = $s0_frg{$identifier};
  $stop = $s1_frg{$identifier};

  # correct positions according to offset
  $start += $offset{$identifier};
  $stop += $offset{$identifier};

  $masked_length = $stop - $start + 1;
  $length = length($sequence);

  # initial fragment
  $masked_sequence = substr($sequence, 0, $start);

  # masked fragment
  for ($i = 0; $i < $masked_length; $i++) {
    $masked_sequence .= 'X';
  }
  # last fragment
  $masked_sequence .= 
    substr($sequence, $stop + 1, $length - $stop);
  
  # if failing to mask, then do print sequence
  if ($masked_sequence ne '') {
    print MASK ">$identifier\n";
    print MASK "$masked_sequence\n";
  }



  # print info about the hit

  # get back modified identifiers
  $original_id = $identifier;
  $original_id =~ s/\_\_/\./g;

  $hit_fragment = substr($sequence, $start, $masked_length);
#  print REP_PAR "$original_id\t$cycle\t$e_value{$identifier}\t";
# Version 2.0 see below

# version 3.4: Print it only in search mode
  if ($mode eq 'search') {
    print "$original_id\t$prot_length{$identifier}\t$cycle\t$e_value{$identifier}\t";
    print "$enext_value{$identifier}\t$score{$identifier}\t";
    print "$masked_length\t$p0_frg{$identifier}\t$p1_frg{$identifier}\t$start\t";
    print "$stop\t$profile{$identifier}\t$hit{$identifier}\n";
  }

}



# delete temporal files

system(`rm $MASK`);
system(`rm $FRG`);
system(`rm $NEG_IN`);
system(`rm $POS_IN`);
system(`rm $ALN`);
system(`rm $HIT`);
system(`rm $FRG_ALN`);
system(`rm $DB`);
system(`rm $ADD`);
system(`rm $DB_ADD`);
system(`rm $FRG_ADD`);
system(`rm $TMP`);

