#!/usr/pub/bin/perl5.00404 
# select.perl
#
# Selects repeats from the parameter file
#
# There is an E-value threshold that is applied to the EVD(n).
# Successive repeats are examined as far as EVD(n+1) is above the
# threshold, but they need to be validated by a later EVD(n).
# When a repeat does not fulfill any of the two conditions the
# search stops and repeats pending validation are discarded.
#
# Use as: select.perl parameter_file e_threshold enext_threshold min_repeats max_prot_length
# e.g.: select.perl wd_neg.par 1e-6 1e-6 4 10000
#
# Version 2.0 March 99: repeater now includes one column (2nd) with
#            the length of the protein. The ouptut must be red,
#            and there is a new possible threshold on maximum protein 
#            length.
#
# Version 2.1 Oct 2000: added option -sw [repeat name]
#            Makes output in swissprot format 
#
# Version 2.2 6th Nov 2000: in sw output no numeral when only one 
#            repeat is reported.
#
# Version 2.3 13 May 2002: Avoid producing overlapping ranges for the 
#	     same kind of repeat (as requested by Amos Bairoch).
#
# Version 2.4 28 May 2002: Correct a bug introduced. Wrong fragments
#           when the input par file contains multiple sequences.
#
# Miguel Andrade  Oct 98
# Repeats project
#####################################################################


# constants

$e_threshold = 1e-6;
$enext_threshold = 1e-6;
$min_repeats = 4;
$max_prot_length = 10000;



if ($#ARGV < 4) {
  die "Use as: select.perl parameter_file e_threshold enext_threshold min_repeats max_prot_length [-sw repeat name]\n";
}


$PAR = $ARGV[0];
$e_threshold = $ARGV[1];
$enext_threshold = $ARGV[2];
$min_repeats = $ARGV[3];
$max_prot_length = $ARGV[4];
open(PAR, $PAR) or die "parameter file not found\n";

$output = 'rdb';
if ($ARGV[5] eq '-sw'){
  $repeat_name = $ARGV[6];
  $output = 'sw';
} 

if ($output eq 'rdb'){
  # print rdb header
  print "#rdb generated by select.perl\n";
  print "id\torder\t";
  print "long_id\tevalue\tenext\tscore\t";
  print "seq_start\tseq_end\tfragment\n";
  print "60S\t2N\t";
  print "60S\t10N\t10N\t3N\t";
  print "3N\t3N\t100S\n";
}

# read data

$counter = 0;
$skip_header = 2;
while(<PAR>) {
  if (/^\#/) {
    next;
  }
  if ($skip_header > 0) {
    $skip_header--;
    next;
  }
  chop;

  @words = split(/\t/);
  
  if ($words[0] =~ /\|/) {
    # identifier is a nrdb like
    @words_id = split(/\|/,$words[0]);
    $id = pop(@words_id);
  } else {
    $id = $words[0];
  }

  $prot_length = $words[1];
  # if the length of the protein exceeds max_prot_length
  # then do not consider the repeat
  if ($prot_length > $max_prot_length) {
    $terminate{$id} = 1;
  }

  $cycle = $words[2];
  
  if (($cycle == 1) or (! exists $terminate{$id})) {
    # check this repeat if the sequence of repeats still 
    # under the thresholds

    $e = $words[3];
    $enext = $words[4];
    $long_id = '';  

    # apply threshold to EVD(n)

    if (($e < $e_threshold) and ($cycle >= $min_repeats)) {
    
      # validate this repeat, 
      # if the E-value is under the threshold    

      $long_id = $id.'/'.$cycle;
      $s0{$long_id} = $words[9];

      $counter++;

      # and all previous pending ones
      $i = $cycle - 1;

      while ($i > 0) {
	$back_id = $id.'/'.$i;
	if (! exists $s0{$back_id}) {
          # validate this pending repeat
	  $s0{$back_id} = $s0_pending{$back_id};
	  $counter++;

	  $i--;
	} else {
	  # terminate
	  $i = 0;
	}
      }

    } elsif (($e < $e_threshold) or ($enext < $enext_threshold)) {

      # assign repeat to pending-state

      $long_id = $id.'/'.$cycle;
      
      $s0_pending{$long_id} = $words[9];
    }


    if ($long_id ne '') {
      
      # get data for this repeat (either it has been validated or pends)

      $hit{$long_id} = $words[12];
      $s1{$long_id} = $words[10];
      $evalue{$long_id} = $words[3];
      $enext{$long_id} = $words[4];
      $score{$long_id} = $words[5];
      
      ($id{$long_id}, $order{$long_id}) = split(/\//, $long_id);


    } else {
      
      # the repeat was neither validated nor pends.
      # terminate the repeat analysis for this sequence
      $terminate{$id} = 1;

    }

  }

}
close(PAR);


# version 2.3
# Clean overlappings
# vesion 2.4
# But separately for each protein!

foreach $long_id(keys %s0){
  $id_list{$id{$long_id}} = 1;
}


foreach $id1(sort keys %id_list){
  undef %t0;
  undef %t1;

  foreach $long_id(keys %s0){

    if ($id{$long_id} eq $id1){
      $t0{$long_id} = $s0{$long_id};
      $t1{$long_id} = $s1{$long_id};
    }
  }

  $previous = 0;
  foreach $long_id(sort byposition keys %t0) {
    $overlap =  $previous - $t0{$long_id} + 1;
    if ($overlap > 0){
      $t0{$long_id} += $overlap;
      $s0{$long_id} += $overlap;
      for ($i = 0; $i < $overlap; $i++){
        $hit{$long_id} =~ s/^.//;
      }  
    }
    $previous = $t1{$long_id};
  }
}	



if ($output eq 'rdb'){
  # unsorted rdb output
  foreach $long_id (keys %s0) {
    print "$id{$long_id}\t$order{$long_id}\t";
    print "$long_id\t$evalue{$long_id}\t$enext{$long_id}\t$score{$long_id}\t";
    print "$s0{$long_id}\t$s1{$long_id}\t$hit{$long_id}\n";
  }
} else {
  # swissprot feature output
  # sort output according to protein name and start of repeat

  
  # annotate the repeat data
  foreach $long_id(keys %s0) {
    $start_list{$id{$long_id}}.= $s0{$long_id}.';';
    $stop_list{$id{$long_id}} .= $s1{$long_id}.';';
  }

  # write lines ordered by identifier
  foreach $id(keys %start_list) {
    print "$repeat_name REPEATS FOR $id\n";

    # reorder the rstart/stop list
    @start = split(/\;/,$start_list{$id});
    @ordered_starts = sort numerically @start;

    @stop = split(/\;/,$stop_list{$id});
    @ordered_stops = sort numerically @stop;

    $a = scalar(@ordered_starts);


    for ($i = 0; $i < $a; $i++){
      $j = $i + 1;

      print "FT   REPEAT";
      $spaces = 9 - length($ordered_starts[$i]);
      for ($k = 0; $k < $spaces; $k++){
	print " ";
      }
      print $ordered_starts[$i];
      $spaces = 7 - length($ordered_stops[$i]);
      for ($k = 0; $k < $spaces; $k++){
	print " ";
      }
      print $ordered_stops[$i];
      # version 2.1
      # print numeral only if more than 1 repeat
      if ($a > 1){
	print "       $repeat_name $j.\n";
      } else {
	print "       $repeat_name.\n";
      }
    }


  }

}


sub numerically { $a <=> $b; }

sub byposition
  {
    $t0{$a} <=> $t0{$b}
  }
